from datetime import datetime
from datetime import timedelta

import yfinance

from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import UniqueConstraint
from sqlalchemy import engine
from sqlalchemy import text
from sqlalchemy.orm import declarative_base
from sqlalchemy.types import Numeric

from foreverbull.pb import pb_utils
from foreverbull.pb.foreverbull.backtest import backtest_pb2
from foreverbull.pb.foreverbull.finance import finance_pb2


Base = declarative_base()


class Asset(Base):
    __tablename__ = "asset"
    symbol = Column("symbol", String(), primary_key=True)
    name = Column("name", String())


class OHLC(Base):
    __tablename__ = "ohlc"
    id = Column(Integer, primary_key=True)
    symbol = Column(String())
    open = Column(Numeric())
    high = Column(Numeric())
    low = Column(Numeric())
    close = Column(Numeric())
    volume = Column(Integer())
    time = Column(DateTime())

    __table_args__ = (UniqueConstraint("symbol", "time", name="symbol_time_uc"),)


def verify(database: engine.Engine, backtest: backtest_pb2.Backtest):
    with database.connect() as conn:
        for symbol in backtest.symbols:
            result = conn.execute(
                text("SELECT min(time), max(time) FROM ohlc WHERE symbol = :symbol"),
                {"symbol": symbol},
            )
            res = result.fetchone()
            if res is None:
                return False
            start, end = res
            if start is None or end is None:
                return False
            if start.date() != pb_utils.from_proto_date_to_pydate(
                backtest.start_date
            ) or end.date() != pb_utils.from_proto_date_to_pydate(backtest.end_date):
                return False
        return True


def populate(database: engine.Engine, backtest: backtest_pb2.Backtest):
    with database.connect() as conn:
        for symbol in backtest.symbols:
            feed = yfinance.Ticker(symbol)
            info = feed.info
            asset = finance_pb2.Asset(
                symbol=info["symbol"],
                name=info["longName"],
            )
            conn.execute(
                text(
                    """INSERT INTO asset (symbol, name)
                    VALUES (:symbol, :name) ON CONFLICT DO NOTHING"""
                ),
                {"symbol": asset.symbol, "name": asset.name},
            )
            data = feed.history(
                start=pb_utils.from_proto_date_to_pydate(backtest.start_date),
                end=pb_utils.from_proto_date_to_pydate(backtest.end_date) + timedelta(days=1),
            )
            for idx, row in data.iterrows():
                time = datetime(idx.year, idx.month, idx.day, idx.hour, idx.minute, idx.second)  # type: ignore
                conn.execute(
                    text(
                        """INSERT INTO ohlc (symbol, open, high, low, close, volume, time)
                        VALUES (:symbol, :open, :high, :low, :close, :volume, :time) ON CONFLICT DO NOTHING"""
                    ),
                    {
                        "symbol": symbol,
                        "open": row.Open,
                        "high": row.High,
                        "low": row.Low,
                        "close": row.Close,
                        "volume": int(row.Volume),
                        "time": time,
                    },
                )
        conn.commit()



import pytest

from foreverbull_testing.data import Asset
from foreverbull_testing.data import Assets


@pytest.fixture(scope="session")
def loaded_connection(fb_database):
    engine, ensure_data = fb_database
    ensure_data(start="2023-01-01", end="2023-02-01", symbols=["AAPL", "MSFT"])
    with engine.connect() as conn:
        yield conn


def test_asset(loaded_connection):
    a = Asset(loaded_connection, "2023-01-01", "2023-02-01", "AAPL")
    assert a.symbol == "AAPL"
    assert a.stock_data is not None


def test_asset_metrics(loaded_connection):
    a = Asset(loaded_connection, "2023-01-01", "2023-02-01", "AAPL")
    a.set_metric("test", 1)
    assert a.get_metric("test") == 1
    assert a.get_metric("test2") is None
    assert a.metrics == {"test": 1}


def test_assets(loaded_connection):
    a = Assets(loaded_connection, "2023-01-01", "2023-02-01", ["AAPL", "MSFT"])
    assert a.symbols == ["AAPL", "MSFT"]
    assert a.stock_data is not None


def test_assets_metrics(loaded_connection):
    a = Assets(loaded_connection, "2023-01-01", "2023-02-01", ["AAPL", "MSFT"])
    a.set_metrics("test", {"AAPL": 1, "MSFT": 2})
    assert a.get_metrics("test").to_dict() == {"AAPL": 1, "MSFT": 2}
    assert a.get_metrics("test2").to_dict() == {}
    assert a.metrics == {"test": {"AAPL": 1, "MSFT": 2}}
